// SPDX-License-Identifier: MIT
// Copyright (c) 2018-2024, Advanced Micro Devices, Inc. All rights reserved.

#include "add_rmsnorm2d_rdquant_fwd.inc"

int main() { return run_add_rmsnorm2d_rdquant_combinations("fp16"); }
