// SPDX-License-Identifier: MIT
// Copyright (c) 2025, Advanced Micro Devices, Inc. All rights reserved.

#include "test_run_gemm_aquant_example.inc"

int main() { return run_gemm_combinations("bf8"); }
