// SPDX-License-Identifier: MIT
// Copyright (c) 2018-2024, Advanced Micro Devices, Inc. All rights reserved.

#pragma once
#include <stdint.h>

namespace ck_tile {

struct null_type
{
};

} // namespace ck_tile
