// Copyright (c) 2024, Jay Shah, Ganesh Bikshandi, Ying Zhang, Vijay Thakkar, Pradeep Ramani, Tri Dao.
// Splitting the different template instantiations to different files to speed up compilation.
// This file is auto-generated. See "generate_kernels.py"

#include "flash_fwd_hdim256_bf16_paged_split_sm80.cu"
#include "flash_fwd_hdim256_bf16_paged_split_softcap_sm80.cu"