// SPDX-License-Identifier: MIT
// Copyright (c) 2018-2025, Advanced Micro Devices, Inc. All rights reserved.

#include "layernorm2d_fwd.inc"

int main() { return run_layernorm2d_fwd_combinations("bf16"); }
